/*
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Model of a hardware interface
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <assert.h>
#include <string>

#include <hw/assertion.hpp>
#include <hw/component.hpp>
#include <hw/matcher.hpp>
#include <hw/message.hpp>
#include <hw/process.hpp>

#include <algorithm>
#include <ostream>
#include <vector>

/** @def CPPTESK_CAPTURE_IFACE
 * Captures an interface.
 */
#define CPPTESK_CAPTURE_IFACE(iface) \
    CPPTESK_CAPTURE_COMPONENT(iface)

/** @def CPPTESK_RELEASE_IFACE
 * Releases an interface.
 */
#define CPPTESK_RELEASE_IFACE(iface) \
    CPPTESK_RELEASE_COMPONENT(iface)

#ifdef HW_SHORT_NAMES
    #define CAPTURE_IFACE(iface) CPPTESK_CAPTURE_IFACE(iface)
    #define RELEASE_IFACE(iface) CPPTESK_RELEASE_IFACE(iface)
#endif // HW_SHORT_NAMES

namespace cpptesk {
namespace hw {

class Interface: public Component
{
public:
    typedef enum { INPUT = 0x01, OUTPUT = 0x02, INOUT = INPUT | OUTPUT } inout_t;

    Interface(inout_t type = INOUT, const std::string &name = "<default>"):
        Component(name), message(NULL), type(type), matcher(NULL), requestCount(0), taken(false) {}

    virtual ~Interface() {}

    void registerReactionMatcher(ReactionMatcher &matcher)
    {
        this->matcher = &matcher;
    }

    bool isInput()  const { return (type & INPUT)  != 0; }
    bool isOutput() const { return (type & OUTPUT) != 0; }

    // Basically for input interfaces
    virtual void request(const Process &process, const Message &message)
    {
        Component::request(process);
        this->message = (Message*)&message;
    }

    virtual void release()
    {
        Component::release();
        message = NULL;
    }

    virtual Message* getMessage() const
    {
        return message;
    }

    // Basically for output interfaces
    virtual void appendRequest(const Process &process, const Message &message)
    {
        if(matcher != NULL)
            { matcher->appendRequest(process, message); }

        requestCount++;
    }

    virtual bool chooseRequest(const Process &process, const Message &message) const
    {
        if(matcher != NULL)
            { return matcher->chooseRequest(process, message); }

        return true;
    }

    virtual void removeRequest(const Process &process, const Message &message)
    {
        if(matcher != NULL)
            { matcher->removeRequest(process, message); }

        requestCount--;
        taken = false;
    }

    virtual int requestNumber() const
    {
        return requestCount;
    }

    /// Returns false if a reaction has been taken by someone else.
    virtual bool takeReaction()
    {
        CPPTESK_ASSERTION(requestCount > 0, "Reaction could not be taken");

        if(taken) { return false; }
        return taken = true;
    }

    virtual bool isReactionTaken() const
    {
        return taken;
    }

protected:
    /// Current input message.
    Message *message;

    /// Type of the interface.
    inout_t type;

    /// Reaction matcher.
    ReactionMatcher *matcher;

    /// Number of requests to take a reaction.
    int requestCount;

    /// Flag that shows if a reaction is taken or not.
    bool taken;
};

class InputInterface: public Interface
{
public:
    InputInterface(const std::string &name = "<default>"): Interface(INPUT, name) {}
    virtual ~InputInterface() {}
};

class OutputInterface: public Interface
{
public:
    OutputInterface(const std::string &name = "<default>"): Interface(OUTPUT, name) {}
    virtual ~OutputInterface() {}
};

class InoutInterface: public Interface
{
public:
    InoutInterface(const std::string &name = "<default>"): Interface(INOUT, name) {}
    virtual ~InoutInterface() {}
};

inline std::ostream& operator << (std::ostream &out, const Interface &iface)
{
	return out << iface.getName();
}

}} // namespace cpptesk::hw

