/*
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Reaction arbiters
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/message.hpp>
#include <hw/process.hpp>

#include <algorithm>
#include <map>
#include <vector>

namespace cpptesk {
namespace hw {

class ReactionMatcher
{
public:
    ReactionMatcher() {}
    virtual ~ReactionMatcher() {}

    virtual void appendRequest(const Process &process, const Message &message) {}
    virtual bool chooseRequest(const Process &process, const Message &message) const { return true; }
    virtual void removeRequest(const Process &process, const Message &message) {}
};

class FIFOReactionMatcher: public ReactionMatcher
{
public:
    FIFOReactionMatcher() {}
    virtual ~FIFOReactionMatcher() {}

    virtual void appendRequest(const Process &process, const Message &message)
    {
        requests.push_back((Process*)&process);
    }

    virtual bool chooseRequest(const Process &process, const Message &message) const
    {
        CPPTESK_ASSERTION(!requests.empty(), "There are no requests to choose");

        return requests[0] == &process;
    }

    virtual void removeRequest(const Process &process, const Message &message)
    {
        CPPTESK_ASSERTION(!requests.empty(), "There are no requests to remove");

        for(std::vector<Process*>::iterator i = requests.begin(); i != requests.end(); i++)
            { if(*i == &process) { requests.erase(i); break; } }
    }

protected:
    /// Queue of reaction requests.
    std::vector<Process*> requests;
};

class PriorityReactionMatcher: public ReactionMatcher
{
public:
    PriorityReactionMatcher() {}
    virtual ~PriorityReactionMatcher() {}

    virtual void appendRequest(const Process &process, const Message &message)
    {
        requests.push_back((Process*)&process);

#ifdef CPPTESK_DEBUG_HW
        printf("PriorityReactionMatcher::appendRequest: pid=%d, priority=%d\n", process.getPID(), process.getPriority());
#endif

        sort(requests.begin(), requests.end(), greaterProcessPriority);
    }

    virtual bool chooseRequest(const Process &process, const Message &message) const
    {
        CPPTESK_ASSERTION(!requests.empty(), "There are no requests to choose");

#ifdef CPPTESK_DEBUG_HW
        for(std::vector<Process*>::const_iterator i = requests.begin(); i != requests.end(); i++)
        {
            printf("PriorityReactionMatcher::chooseRequest: pid=%d, priority=%d\n", (*i)->pid, (*i)->priority);
        }
#endif

        return requests[0] == &process;
    }

    virtual void removeRequest(const Process &process, const Message &message)
    {
        CPPTESK_ASSERTION(!requests.empty(), "There are no requests to remove");

#ifdef CPPTESK_DEBUG_HW
        printf("PriorityReactionMatcher::removeRequest: pid=%d\n", process.getPID());
#endif

        for(std::vector<Process*>::iterator i = requests.begin(); i != requests.end(); i++)
            { if(*i == &process) { requests.erase(i); break; } }
    }

protected:
    /// Queue of reaction requests.
    std::vector<Process*> requests;
};

class InterfaceOrderingReactionMatcher: public ReactionMatcher
{
public:
    InterfaceOrderingReactionMatcher() {}
    virtual ~InterfaceOrderingReactionMatcher() {}

    virtual void appendRequest(const Process &process, const Message &message)
    {
        std::vector<Process*> &processes = requests[process.input_iface];
        processes.push_back((Process*)&process);

#ifdef CPPTESK_DEBUG_HW
        printf("InterfaceOrderingReactionMatcher::appendRequest: pid=%d, priority=%d\n", process.getPID(), process.getPriority());
#endif

        //sort(requests_vector.begin(), requests_vector.end(), ctesk::greaterProcessPriority);
    }

    virtual bool chooseRequest(const Process &process, const Message &message) const
    {
        const std::vector<Process*> processes = requests.find(process.input_iface)->second;

        CPPTESK_ASSERTION(!processes.empty(), "There are no requests to choose");
        
#ifdef CPPTESK_DEBUG_HW
        for(std::vector<Process*>::const_iterator i = processes.begin(); i != processes.end(); i++)
        {
            printf("InterfaceOrderingReactionMatcher::chooseRequest: pid=%d, priority=%d\n", (*i)->pid, (*i)->priority);
        }
#endif

        return processes[0] == &process;
    }

    virtual void removeRequest(const Process &process, const Message &message)
    {
        std::vector<Process*> &processes = requests[process.input_iface];
        CPPTESK_ASSERTION(!processes.empty(), "There are no requests to remove");

#ifdef CPPTESK_DEBUG_HW
        printf("InterfaceOrderingReactionMatcher::removeRequest: pid=%d\n", process.getPID());
#endif

        for(std::vector<Process*>::iterator i = processes.begin(); i != processes.end(); i++)
        { if(*i == &process) { processes.erase(i); break; } }

    }

protected:
    typedef std::vector<Process*> requestvector;
    std::map<Interface*, requestvector> requests;
};

}} // namespace cpptesk::hw

