/*
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Model of a message
///////////////////////////////////////////////////////////////////////////////

#include <hw/assertion.hpp>
#include <hw/message.hpp>

namespace cpptesk {
namespace hw {

void Message::copyFields(const Message &rhs)
{
    for(size_t i = 0; i < _fields.size(); i++)
    {
        CPPTESK_ASSERTION(i < rhs._fields.size(), "Different number of fields");

        field_t lhs_field = _fields[i];
        field_t rhs_field = rhs._fields[i];

        this->*lhs_field = rhs.*rhs_field;
    }
}

Message::Message(const Message &rhs):
    valid(rhs.valid), explanation(rhs.explanation), sysdata(rhs.sysdata)
{
    CPPTESK_ASSERTION(_fields.empty(), "List of fields is not empty");

    _fields = rhs._fields;

    copyFields(rhs);
}

Message& Message::operator = (const Message &rhs)
{
    if(&rhs != this)
    {
        valid = rhs.valid;
        explanation = rhs.explanation;
        sysdata = rhs.sysdata;

        copyFields(rhs);
    }

    return *this;
}

/// Compares a message with another one and returns non-empty string if they are not equal.
CPPTESK_DEFINE_COMPARATOR(Message)
{
    // If this message (which is a reference one) is invalid,
    // then comparison is not carried out (i.e. invalid = unknown)
    if(!isValid())
        { return ""; }

    for(size_t i = 0; i < _fields.size(); i++)
    {
        CPPTESK_ASSERTION(i < message._fields.size(), "Different number of fields");

        field_t lhs_field = _fields[i];
        field_t rhs_field = message._fields[i];

        std::string error = (const_cast<Message*>(this)->*lhs_field).compare(message.*rhs_field);

        if(!error.empty()) { return error; }
    }

    return "";
}

/// Returns a string representation of a message.
std::string Message::toString() const
{
    std::stringstream buffer;

    for(size_t i = 0; i < _fields.size(); i++)
    {
        field_t field = _fields[i];
        // Convert into string
        buffer << (this->*field).getName() << "(" << (this->*field).toString() << ") ";
    }

    return buffer.str();
}

/// Randomizes a message data.
void Message::randomize()
{
    for(size_t i = 0; i < _fields.size(); i++)
    {
        field_t field = _fields[i];
        // Randomize a field
        (this->*field).randomize();
    }
}

EmptyMessage CPPTESK_DEFAULT_MESSAGE;

}} // namespace cpptesk::hw

