/*
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Model of a message
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/clone.hpp>
#include <hw/errors.hpp>
#include <hw/process.hpp>
#include <hw/random.hpp>
#include <hw/time.hpp>

#include <stdint.h>
#include <string.h>

#include <sstream>
#include <ostream>
#include <string>
#include <vector>

////////////////////////////////////////////////////////////////////////////////
// Declaration of Message Classes
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_MESSAGE
 * Declares a model message.
 */
#define CPPTESK_MESSAGE(UserMessage) \
    struct UserMessage: public ::cpptesk::hw::Message

#ifdef HW_SHORT_NAMES
    #define MESSAGE(UserMessage) CPPTESK_MESSAGE(UserMessage)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Declaration of Message Fields
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_DECLARE_MASKED_FIELD
 * Declares a masked field.
 */
#define CPPTESK_DECLARE_MASKED_FIELD(field_name, field_width, field_mask) \
    ::cpptesk::hw::Field<field_width, field_mask> field_name; \
    inline uint64_t get_##field_name() const { return field_name.get(); } \
    inline void set_##field_name(uint64_t value) { field_name.set(value); }

#define __FIELD_MASK(field_width) \
    ((field_width) == 64 ? 0xffffFFFFffffFFFFull : (1ull << (field_width)) - 1)

/** @def CPPTESK_DECLARE_FIELD
 * Declares a field.
 */
#define CPPTESK_DECLARE_FIELD(field_name, field_width) \
    CPPTESK_DECLARE_MASKED_FIELD(field_name, field_width, __FIELD_MASK(field_width))

/** @def CPPTESK_DECLARE_BIT
 * Declares a single-bit field.
 */
#define CPPTESK_DECLARE_BIT(field_name) \
    CPPTESK_DECLARE_FIELD(field_name, 1)

#ifdef HW_SHORT_NAMES
    #define DECLARE_MASKED_FIELD(field_name, field_width, field_mask) CPPTESK_DECLARE_MASKED_FIELD(field_name, field_width, field_mask)
    #define DECLARE_FIELD(field_name, field_width)                    CPPTESK_DECLARE_FIELD(field_name, field_width)
    #define DECLARE_BIT(field_name)                                   CPPTESK_DECLARE_BIT(field_name)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Declaration and Definition of a Comparator
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_DECLARE_COMPARATOR
 * Declares a comparator in a message class.
 */
#define CPPTESK_DECLARE_COMPARATOR(UserMessage) \
    /* Overloads the basic comparator */ \
    virtual std::string compare(const Message &message) const \
        { return compare(dynamic_cast<const UserMessage&>(message)); } \
    virtual std::string compare(const UserMessage &message) const

/** @def CPPTESK_DEFINE_COMPARATOR
 * Defines a comparator.
 */
#define CPPTESK_DEFINE_COMPARATOR(UserMessage) \
    std::string UserMessage::compare(const UserMessage &message) const

#ifdef HW_SHORT_NAMES
    #define DECLARE_COMPARATOR(UserMessage) CPPTESK_DECLARE_COMPARATOR(UserMessage)
    #define DEFINE_COMPARATOR(UserMessage)  CPPTESK_DEFINE_COMPARATOR(UserMessage)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Requirements Traceability
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_SET_REQUIREMENT
 * Sets a requirement (some string explaning a reason for sending or receiving a message).
 */
#define CPPTESK_SET_REQUIREMENT(msg, req) \
    do { \
        std::stringstream __buffer; \
        __buffer << req \
            << " at cycle [" << time() << ".." << time() + getReactionTimeout() << "] (" \
            << __FILE__ << ", " << __LINE__ << ")"; \
        (msg).setExplanation(__buffer.str()); \
    } while(false)

/** @def CPPTESK_SET_REQUIREMENT
 * Gets a message's requirement.
 */
#define CPPTESK_GET_REQUIREMENT(msg) \
    ((msg).getExplanation())

/** @def CPPTESK_SET_COMMENT
 * Alias of CPPTESK_SET_REQUIREMENT
 */
#define CPPTESK_SET_COMMENT(msg, req) CPPTESK_SET_REQUIREMENT(msg, req)

/** @def CPPTESK_GET_COMMENT
 * Alias of CPPTESK_GET_REQUIREMENT
 */
#define CPPTESK_GET_COMMENT(msg) CPPTESK_GET_REQUIREMENT(msg)

#ifdef HW_SHORT_NAMES
    #define SET_REQUIREMENT(msg, req) CPPTESK_SET_REQUIREMENT(msg, req)
    #define GET_REQUIREMENT(msg)      CPPTESK_GET_REQUIREMENT(msg)
    #define SET_COMMENT(msg, req)     CPPTESK_SET_COMMENT(msg, req)
    #define GET_COMMENT(msg)          CPPTESK_GET_COMMENT(msg)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Field Registration
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_ADD_FIELD
 * Registers a field in a message class.
 */
#define CPPTESK_ADD_FIELD(full_field_name) \
    registerField(#full_field_name, (Message::field_t)&full_field_name)

#ifdef HW_SHORT_NAMES
    #define ADD_FIELD(full_field_name) CPPTESK_ADD_FIELD(full_field_name)
#endif // HW_SHORT_NAMES

/////////////////////////////////////////////////////////////////////////////////
// Clone Support
/////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_SUPPORT_CLONE(UserMessage) \
    virtual UserMessage* clone() { return new UserMessage(*this); }

#ifdef HW_SHORT_NAMES
    #define SUPPORT_CLONE(UserMessage) CPPTESK_SUPPORT_CLONE(UserMessage)
#endif // HW_SHORT_NAMES

/////////////////////////////////////////////////////////////////////////////////
// Message Transmission Delay & Priority
/////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_MESSAGE_DELAY(msg, delay) \
    (msg).setDelay(delay)

#define CPPTESK_MESSAGE_LOCK(msg, pid) \
    (msg).setLock(pid)

#ifdef HW_SHORT_NAMES
    #define MESSAGE_DELAY(msg, delay) CPPTESK_MESSAGE_DELAY(msg, delay)
    #define MESSAGE_LOCK(msg, pid)    CPPTESK_MESSAGE_LOCK(msg, pid)
#endif

namespace cpptesk {
namespace hw {

inline uint16_t reverse_byte_order(uint16_t value)
{
    return ((value & 0xff) << 8)
         | ((value >> 8) & 0xff);
}

inline uint32_t reverse_byte_order(uint32_t value)
{
    return (uint32_t)reverse_byte_order((uint16_t)((value & 0xffff))) << 16
         | (uint32_t)reverse_byte_order((uint16_t)((value >> 16) & 0xffff));
}

inline uint64_t reverse_byte_order(uint64_t value)
{
    return (uint64_t)reverse_byte_order((uint32_t)((value & 0xffffFFFFul))) << 32
         | (uint64_t)reverse_byte_order((uint32_t)((value >> 32) & 0xffffFFFFul));
}

inline uint64_t get_mask(unsigned width)
{
    return (1ull << width) - 1;
}

inline uint64_t get_mask(unsigned min, unsigned max)
{
    return get_mask((max - min) + 1) << min;
}

inline uint64_t get_field(uint64_t value, unsigned min, unsigned max)
{
    return (value >> min) & get_mask((max - min) + 1);
}

inline uint64_t get_field(uint64_t value, unsigned bit)
{
    return get_field(value, bit, bit);
}

inline void set_field(uint64_t &value, unsigned min, unsigned max, uint64_t field)
{
    uint64_t mask = get_mask(min, max);
    value &= ~mask; value |= (field << min) & mask;
}

inline void set_field(uint64_t &value, unsigned bit, uint64_t field)
{
    set_field(value, bit, bit, field);
}

typedef enum interlock_t
{
    LOCK_NONE       = 0x00,
    LOCK_HOLD       = 0x01,
    LOCK_DELAY      = 0x02,
    LOCK_HOLD_DELAY = LOCK_HOLD | LOCK_DELAY
} interlock_t;

////////////////////////////////////////////////////////////////////////////////
// Message Field
////////////////////////////////////////////////////////////////////////////////

template<size_t field_width = 64, uint64_t field_mask = 0xffffFFFFffffFFFFull>
class Field
{
public:
    Field(): _data() {}
    virtual ~Field() {}

    Field(const Field &rhs) { _data = rhs._data; }

    Field& operator = (const Field &rhs)    { _data = rhs._data; return *this; }
    Field& operator = (const uint64_t &rhs) { _data = rhs; return *this; }

    operator uint64_t () { return _data; }

    inline uint64_t get() const { return _data; }
    inline void set(uint64_t value) { _data = value & get_mask(field_width) & field_mask; }

    inline std::string getName() const { return _name; }
    inline void setName(const std::string &name) { _name = name; }

    virtual std::string compare(const Field &rhs)
    {
        std::stringstream buffer;

        if(get() != rhs.get())
        {
            buffer << getName() << "(";
            buffer << "spec=" << toString() << ", ";
            buffer << "impl=" << rhs.toString() << ")";
        }

        return buffer.str();
    }

    virtual std::string toString() const
    {
        std::stringstream buffer;
        buffer << std::hex << _data;
        return buffer.str();
    }

    virtual void randomize() { set(Random::create().rand<uint64_t>(field_width)); }

protected:
    uint64_t _data;
    std::string _name;
};

class Message:
    public Randomizable
  , public Cloneable
{
public:
    Message(): valid(true), explanation("Unknown reason") { memset(&sysdata, 0, sizeof(sysdata)); }
    virtual ~Message() {}

    Message(const Message &rhs);

    Message& operator = (const Message &rhs);

#ifdef CPPTESK_CLONE_INPUT_MESSAGE
    virtual Message* clone() = 0;
#else
    virtual Message* clone() { return this; }
#endif

    std::string getName() const { return _name; }
    void setName(const std::string &name) { _name = name; }

    void setSystemData(priority_t priority, interlock_t interlock, pid_t pid, cycle_t delay = 0)
    {
        sysdata.priority = priority;
        sysdata.interlock = interlock;
        sysdata.pid = pid;
        sysdata.delay = delay;
    }

	void setPriority(priority_t priority) { sysdata.priority = priority; }
	void setDelay(cycle_t delay)          { sysdata.interlock = LOCK_DELAY; sysdata.delay = delay; }
	void setHold(pid_t pid)               { sysdata.interlock = LOCK_HOLD; sysdata.pid = pid; }

    void setExplanation(const std::string &explanation) { this->explanation = explanation; }
    const std::string& getExplanation() const { return explanation; }

    priority_t getPriority() const { return sysdata.priority; }
    interlock_t getInterlock() const { return sysdata.interlock; }
    pid_t getPID() const { return sysdata.pid; }
    int getDelay() const { return sysdata.delay; }

    void invalidate() { this->valid = false; }
    void setValid(bool valid = true) { this->valid = valid; }
    bool isValid() const { return valid; }

    /// Compares a message with another one and returns non-empty string if they are not equal.
    virtual std::string compare(const Message &message) const;

    /// Returns a string representation of a message.
    virtual std::string toString() const;
    /// Randomizes a message data.
    virtual void randomize();

protected:
    std::string _name;

    /// Registry of message fields.
    typedef Field<> Message::*field_t;
    typedef std::vector<field_t> fields_list_t;
    fields_list_t _fields;

    void registerField(const std::string &name, field_t field) { (this->*field).setName(name), _fields.push_back(field); }
    void copyFields(const Message &rhs);

    bool valid;

    std::string explanation;

    struct {
        priority_t priority;
        interlock_t interlock;
        pid_t pid;
        cycle_t delay;
    } sysdata;
};

/**
 * Empty message being used as an input message for input-independent processes,
 * like reaction listeners and other infinite loops.
 */
CPPTESK_MESSAGE(EmptyMessage)
{
    EmptyMessage(): Message() {}
    EmptyMessage(const EmptyMessage &rhs): Message(rhs) {}

    virtual ~EmptyMessage() {}

    CPPTESK_SUPPORT_CLONE(EmptyMessage);
};

extern EmptyMessage CPPTESK_DEFAULT_MESSAGE;

inline std::ostream& operator << (std::ostream& out, const Message &msg)
{
    return out << msg.toString();
}

}} // namespace cpptesk::hw
