/*
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Model of an operation execution process
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/time.hpp>

#include <stdint.h>
#include <ucontext.h>

namespace cpptesk {
namespace hw {

class Interface;
class Message;

/** @typedef pid_t
 * Process identifier.
 */
typedef int pid_t;

typedef enum priority_t
{
    NORMAL_PRIORITY  = 0x10,
    LOWEST_PRIORITY  = 0x00,
    HIGHEST_PRIORITY = 0xff,
    PRIORITY_000     = LOWEST_PRIORITY +  0,
    PRIORITY_001     = LOWEST_PRIORITY +  1,
    PRIORITY_002     = LOWEST_PRIORITY +  2,
    PRIORITY_003     = LOWEST_PRIORITY +  3,
    PRIORITY_004     = LOWEST_PRIORITY +  4,
    PRIORITY_005     = LOWEST_PRIORITY +  5,
    PRIORITY_006     = LOWEST_PRIORITY +  6,
    PRIORITY_007     = LOWEST_PRIORITY +  7,
    PRIORITY_008     = LOWEST_PRIORITY +  8,
    PRIORITY_009     = LOWEST_PRIORITY +  9,
    PRIORITY_010     = LOWEST_PRIORITY + 10,
    PRIORITY_011     = LOWEST_PRIORITY + 11,
    PRIORITY_012     = LOWEST_PRIORITY + 12,
    PRIORITY_013     = LOWEST_PRIORITY + 13,
    PRIORITY_014     = LOWEST_PRIORITY + 14,
    PRIORITY_015     = LOWEST_PRIORITY + 15,
    PRIORITY_016     = LOWEST_PRIORITY + 16,
    PRIORITY_017     = LOWEST_PRIORITY + 17,
    PRIORITY_018     = LOWEST_PRIORITY + 18,
    PRIORITY_019     = LOWEST_PRIORITY + 19,
    PRIORITY_020     = LOWEST_PRIORITY + 20,
    PRIORITY_021     = LOWEST_PRIORITY + 21,
    PRIORITY_022     = LOWEST_PRIORITY + 22,
    PRIORITY_023     = LOWEST_PRIORITY + 23,
    PRIORITY_024     = LOWEST_PRIORITY + 24,
    PRIORITY_025     = LOWEST_PRIORITY + 25,
    PRIORITY_026     = LOWEST_PRIORITY + 26,
    PRIORITY_027     = LOWEST_PRIORITY + 27,
    PRIORITY_028     = LOWEST_PRIORITY + 28,
    PRIORITY_029     = LOWEST_PRIORITY + 29,
    PRIORITY_030     = LOWEST_PRIORITY + 30,
    PRIORITY_031     = LOWEST_PRIORITY + 31,
    PRIORITY_032     = LOWEST_PRIORITY + 32,
    PRIORITY_033     = LOWEST_PRIORITY + 33,
    PRIORITY_034     = LOWEST_PRIORITY + 34,
    PRIORITY_035     = LOWEST_PRIORITY + 35,
    PRIORITY_036     = LOWEST_PRIORITY + 36,
    PRIORITY_037     = LOWEST_PRIORITY + 37,
    PRIORITY_038     = LOWEST_PRIORITY + 38,
    PRIORITY_039     = LOWEST_PRIORITY + 39,
    PRIORITY_040     = LOWEST_PRIORITY + 40,
    PRIORITY_041     = LOWEST_PRIORITY + 41,
    PRIORITY_042     = LOWEST_PRIORITY + 42,
    PRIORITY_043     = LOWEST_PRIORITY + 43,
    PRIORITY_044     = LOWEST_PRIORITY + 44,
    PRIORITY_045     = LOWEST_PRIORITY + 45,
    PRIORITY_046     = LOWEST_PRIORITY + 46,
    PRIORITY_047     = LOWEST_PRIORITY + 47,
    PRIORITY_048     = LOWEST_PRIORITY + 48,
    PRIORITY_049     = LOWEST_PRIORITY + 49,
    PRIORITY_050     = LOWEST_PRIORITY + 50,
} priority_t;

#define PRIORITY_MASK 0xff

/** @typedef status_t
 * Process status.
 */
typedef enum status_t
{
    /// Process is still executing.
    EXECUTING = -1,
    /// Process execution has been successfully completed.
    COMPLETED =  0,
    /// Precondition violation.
    UNPREDICT =  1,
    /// Postcondition violation.
    VIOLATION =  2,
} status_t;

/** @typedef point_t
 * Synchronization point type.
 */
typedef enum point_t
{
    /// Initial synchronization point (process is not executed).
    START = 0,
    /// Synchronization caused by <code>CYCLE()</code> or <code>DELAY</code>.
    CYCLE = 1,
    /// Synchronization caused by <code>GUARD(condition)</code> or other wait condition.
    GUARD = 2,
} point_t;

class Process
{
public:
    Process(pid_t pid, Interface &iface, Message &message);
    virtual ~Process();

    inline pid_t getPID() const { return pid; }

    inline priority_t getPriority() const { return this->priority; }
    inline void setPriority(priority_t priority) { this->priority = priority; }

    inline bool isStarted() const { return syn_point != START; }
    inline bool isWaiting() const { return syn_point == GUARD; }
    inline bool isActive()  const { return syn_point == CYCLE; }

    ////////////////////////////////////////////////////////////////////////////
    // For internal use only
    ////////////////////////////////////////////////////////////////////////////

    /// Process identifier.
    pid_t pid;
    /// Process priority.
    priority_t priority;
     /// Status of the process execution.
    status_t status;
    /// Process context.
    ucontext_t* ucp;
    /// Synchronization point type.
    point_t syn_point;
    /// Starting time.
    cycle_t clk_start;
    /// Process execution time (number of clock cycles passed from the process start-up until now).
    cycle_t clk_total;
    /// Stage execution time (number of clock cycles passed within the current synchronization point).
    cycle_t clk_count;
    /// Timeout.
    bool timeout;
    /// Line where pre- or post-condition violation occured.
    int line;
    /// File where pre- or post-condition violation occured.
    const char *file;
    /// Input interface.
    Interface *input_iface;
    /// Copy of an input message.
    Message *input_message;
    /// Original message.
    Message *original_message;
};

extern bool greaterProcessPriority(Process *lhs, Process *rhs);
extern bool lessProcessPriority(Process *lhs, Process *rhs);

}} // namespace cpptesk::hw

