/*
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Model of a hardware design's reaction
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/errors.hpp>
#include <hw/message.hpp>
#include <hw/time.hpp>

#include <string>
#include <memory>

namespace cpptesk {
namespace hw {

class Reaction 
{
public:
    Reaction(error_type_t error):
        error(error), type(-1), cycle(-1), iface(), info() {}

    Reaction(error_type_t error, uint32_t type, cycle_t cycle):
        error(error), type(type), cycle(cycle), iface(), info() {}

    Reaction(error_type_t error, uint32_t type, cycle_t cycle, const std::string &iface, const std::string &info):
        error(error), type(type), cycle(cycle), iface(iface), info(info) {}

    Reaction(const Reaction &rhs):
        error(rhs.error), type(rhs.type), cycle(rhs.cycle), iface(rhs.iface), info(rhs.info) {}

    virtual ~Reaction() {}

    /**
     * Returns the error type.
     */
    error_type_t getError() const { return error; }

    /**
     * Returns the reaction's type.
     */
    uint32_t getType() const { return type; }

    /**
     * Returns the cycle when the reaction has occurred.
     */
    uint32_t getCycle() const { return cycle; }

    /**
     * Checks whether the reaction is valid or not.
     */
    bool isValid() const { return info.empty(); }

    /**
     * Returns the interface on which the reaction has been detected.
     */
    const std::string& getInterface() const { return iface; }

    /**
     * Returns the reaction's info.
     */
    const std::string& getInfo() const { return info; }

    /**
     * Returns the reaction's description
     */
    std::string getDescription() const
    {
        std::stringstream buffer;
        buffer << getErrorString(error) << "(" << iface << ", " << std::dec << cycle << "): " << info;
        return buffer.str();
    }

protected:
    error_type_t error;
    /// Type of the reaction.
    uint32_t type;
    /// Cycle when the reaction has occurred.
    cycle_t cycle;
    /// Name of the interface on which the reaction has been detected.
    std::string iface;
    /// Description (if info is not empty, the reaction is incorrect).
    std::string info;
};

}} // namespace cpptesk::hw

