#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <stdarg.h>
#include <sys/socket.h>
#include <stdint.h>
#include <dirent.h>
#include <microhttpd.h>

#include "atl/stringbuffer.h"

#include "mcc/mcc.h"
#include "command.h"

#define COORD_PORT       2010

#define CONTEXT          "context"
#define STATUS           "status"
#define STATUS_OK        " "
#define STATUS_STARTING  "Starting"
#define STATUS_STOPPING  "Stopping"
#define STATUS_KILLING   "Killing"
#define STATUS_SAVED     "Suite Saved"
#define STATUS_LOADED    "Suite Loaded"
#define STATUS_SETTINGS  "Settings Updated"

#define CLUSTERS_ID      "clusters"
#define CLUSTER_ID       "cluster"
#define USERS_ID         "users"
#define SETTINGS_ID      "settings"
#define SUITES_ID        "suites"
#define TESTS_ID         "tests"
#define LINKS_ID         "links"
#define TESTRUNS_ID      "testruns"
#define START_ID         "start"
#define INFO_ID          "info"
#define TEST_ID          "test"
#define STOP_ID          "stop"
#define KILL_ID          "kill"
#define SAVE_ID          "save"
#define LOAD_ID          "load"

#define BUF_SIZE 1024

/*
 * Add status properties to the object
 */
static void add_status( StringBuffer* buf, String* context, String* status, const char* default_status )
{
    const char* separator = NULL;
    switch( size_StringBuffer( r(buf) ) )
    {
    case 0:
        separator = "{";
        break;
    case 1:
        separator = "";
        break;
    default:
        separator = ",";
        break;
    }
    appendString_StringBuffer( r(buf)
                             , format_String( "%s\"" CONTEXT "\":\"%s\",\"" STATUS "\":\"%s\"}"
                                            , separator
                                            , toCharArray_String( r(context) )
                                            , status != NULL ? toCharArray_String( r(status) ) : default_status
                                            )
                             );
}

static String* get_cluster_settings( const char* context )
{
    StringBuffer* res = create_StringBuffer();
    char buf[BUF_SIZE];
    String* message = NULL;
    const char* cluster = coord_getCluster();
    String* cluster_file_name_string = format_String("%s/%s/%s/cluster.json"
                                                    , coord_getConfigDir()
                                                    , CLUSTERS_ID
                                                    , cluster
                                                    );
    const char* cluster_file_name = toCharArray_String( r(cluster_file_name_string) );
    FILE* cluster_file;
    append_StringBuffer( r(res), "{\"" );
    append_StringBuffer( r(res), context );
    append_StringBuffer( r(res), "\":" );
    if( (cluster_file = fopen( cluster_file_name, "r" )) == NULL )
    {
        message = getPerrorString( cluster_file_name );
        goto end;
    }
    while( fgets( buf, BUF_SIZE, cluster_file ) != NULL )
    {
        append_StringBuffer( r(res), buf );
    }
    if( feof( cluster_file ) == 0 )
    {
        clear_StringBuffer( r(res) );
        message = getPerrorString( cluster_file_name );
        goto end;
    }
    fclose( cluster_file );
  end:
    destroy( cluster_file_name_string );
    add_status( r(res), create_String( context ), message, STATUS_OK );
    return toString_Object(res);
}

static String* get_full_path( const char* dir_name )
{
    return format_String("%s/%s"
                        , coord_getConfigDir()
                        , dir_name
                        );
}

static String* get_dir_list( String* dir_name_string, const char* target )
{
    const char* dir_name = toCharArray_String( r(dir_name_string) );
    DIR *dir = opendir( dir_name );
    struct dirent *ent;
    int index = 0;
    StringBuffer* buf  = create_StringBuffer();
    String* status = NULL;

    if( dir == NULL )
    {
        status = format_String( "'%s' not found", dir_name );
        goto end;
    }

    append_StringBuffer( r(buf), "{\"" );
    append_StringBuffer( r(buf), target );
    append_StringBuffer( r(buf), "\":[" );

    /* print all the files and directories within directory */
    while( (ent = readdir( dir ) ) != NULL )
    {
        if( /*ent->d_type == DT_DIR &&*/ ent->d_name[0] != '.' )
        {
            if( index++ > 0 )
            {
                append_StringBuffer( r(buf), "," );
            }
            append_StringBuffer( r(buf), "\"" );
            append_StringBuffer( r(buf), ent->d_name );
            append_StringBuffer( r(buf), "\"" );
        }
    }
    closedir( dir );
    append_StringBuffer( r(buf), "]" );
end:
    add_status( r(buf), create_String( target ), status, STATUS_OK );

    destroy(dir_name_string);

    return toString_Object(buf);
}

static String* save_tests( const char* name, const char* tests, const char* context )
{
    String* message = NULL;
    String* suite_dir_name_string = format_String("%s/%s/%s"
                                                 , coord_getConfigDir()
                                                 , SUITES_ID
                                                 , name
                                                 );
    const char* suite_dir_name = toCharArray_String( r(suite_dir_name_string) );
    String* suite_file_name_string = format_String("%s/%s"
                                                  , suite_dir_name
                                                  , "suite.json"
                                                  );
    const char* suite_file_name = toCharArray_String( r(suite_file_name_string) );
    FILE* suite_file;
    if( mkdir( suite_dir_name, 0777 ) != 0 )
    {
        message = getPerrorString( suite_dir_name );
        goto end;
    }
    if( (suite_file = fopen( suite_file_name, "w" )) == NULL )
    {
        message = getPerrorString( suite_file_name );
        goto end;
    }
    if( fputs( tests, suite_file ) == EOF )
    {
        message = getPerrorString( "writing" );
        goto end;
    }

    fclose( suite_file );
  end:
    destroy( suite_dir_name_string );
    destroy( suite_file_name_string );

    StringBuffer* buf = create_StringBuffer();
    add_status( r(buf), create_String( context ), message, STATUS_SAVED );
    return toString_Object(buf);
}

static String* load_tests( const char* name, const char* context )
{
    StringBuffer* res = create_StringBuffer();
    char buf[BUF_SIZE];
    String* message = NULL;
    String* suite_dir_name_string = format_String("%s/%s/%s"
                                                 , coord_getConfigDir()
                                                 , SUITES_ID
                                                 , name
                                                 );
    const char* suite_dir_name = toCharArray_String( r(suite_dir_name_string) );
    String* suite_file_name_string = format_String("%s/%s"
                                                  , suite_dir_name
                                                  , "suite.json"
                                                  );
    const char* suite_file_name = toCharArray_String( r(suite_file_name_string) );
    FILE* suite_file;
    if( (suite_file = fopen( suite_file_name, "r" )) == NULL )
    {
        message = getPerrorString( suite_file_name );
        goto end;
    }
    append_StringBuffer( r(res), "{\"" );
    append_StringBuffer( r(res), context );
    append_StringBuffer( r(res), "\":" );
    while( fgets( buf, BUF_SIZE, suite_file ) != NULL )
    {
        append_StringBuffer( r(res), buf );
    }
    if( feof( suite_file ) == 0 )
    {
        clear_StringBuffer( r(res) );
        message = getPerrorString( suite_file_name );
        goto end;
    }
    fclose( suite_file );
  end:
    destroy( suite_dir_name_string );
    destroy( suite_file_name_string );

    add_status( r(res), create_String( context ), message, STATUS_LOADED );
    return toString_Object(res);
}

static String* update_settings( const char* username, const char* context, const char* settings, int create )
{
    StringBuffer* res = create_StringBuffer();
    String* message = NULL;
    String* user_dir_name_string = format_String("%s/%s/%s"
                                                , coord_getConfigDir()
                                                , USERS_ID
                                                , username
                                                );
    const char* user_dir_name = toCharArray_String( r(user_dir_name_string) );
    String* user_file_name_string = format_String("%s/%s"
                                                 , user_dir_name
                                                 , "settings.json"
                                                 );
    const char* user_file_name = toCharArray_String( r(user_file_name_string) );
    FILE* user_file;
    if( create && mkdir( user_dir_name, 0777 ) != 0 )
    {
        message = getPerrorString( user_dir_name );
        goto end;
    }
    if( (user_file = fopen( user_file_name, settings != NULL ? "w" : "r" )) == NULL )
    {
        message = getPerrorString( user_file_name );
        goto end;
    }

    if( settings )
    {
        // save
        if( fputs( settings, user_file ) == EOF )
        {
            message = getPerrorString( "writing" );
            goto end;
        }
        append_StringBuffer( r(res), "{\"username\":\"" );
        append_StringBuffer( r(res), username );
        append_StringBuffer( r(res), "\",\"settings\":" );
        append_StringBuffer( r(res), settings );
    } else {
        // load
        char buf[BUF_SIZE];

        append_StringBuffer( r(res), "{\"username\":\"" );
        append_StringBuffer( r(res), username );
        append_StringBuffer( r(res), "\",\"settings\":" );
        while( fgets( buf, BUF_SIZE, user_file ) != NULL )
        {
            append_StringBuffer( r(res), buf );
        }
        if( feof( user_file ) == 0 )
        {
            clear_StringBuffer( r(res) );
            message = getPerrorString( user_file_name );
            goto end;
        }
    }

    fclose( user_file );
  end:
    destroy( user_dir_name_string );
    destroy( user_file_name_string );

    add_status( r(res), create_String( context ), message, STATUS_SETTINGS );
    return toString_Object(res);
}

static const char* get_file( const char* url, int* size )
{
    char* buf;
    FILE* f;
    struct stat st;

    if( strlen( url ) == 0 ) url = "index.html";

    chdir( "www" );
    stat( url, &st );
    *size = st.st_size;
    buf = malloc( *size );
    if( !buf )  return 0;
    f = fopen( url, "r" );
    chdir("..");
    if( !f ) return 0;
    *size = fread( buf, 1, *size, f );
    fclose(f);
    return buf;
}

static int respond( struct MHD_Connection *connection, char* bindata, int size, const char* content_type )
{
    int ret;
    struct MHD_Response *response = MHD_create_response_from_data( size
                                                                 , bindata
                                                                 , MHD_NO
                                                                 , MHD_YES
                                                                 );
    if( content_type )
    {
        MHD_add_response_header( response, MHD_HTTP_HEADER_CONTENT_TYPE, content_type );
    }
    ret = MHD_queue_response( connection, bindata != NULL ? MHD_HTTP_OK : MHD_HTTP_NOT_FOUND, response );
    MHD_destroy_response( response );
    if( content_type == NULL )
    {
        free(bindata);
    }
    return ret;
}

static int process_get_request( struct MHD_Connection *connection, const char *url )
{
    const char* bindata = NULL;
    int size = 0;
    String* data = 0;
    const char* content_type = "application/json";

    if( url[0] != '/' ) return MHD_NO;
    url++;

    if( strcmp( url, CLUSTER_ID ) == 0 )
    {
        const char* context = MHD_lookup_connection_value( connection
                                                         , MHD_GET_ARGUMENT_KIND
                                                         , "context"
                                                         );
        data = get_cluster_settings( context );
    } else
    if( strcmp( url, USERS_ID ) == 0 )
    {
        data = get_dir_list( get_full_path(USERS_ID), USERS_ID );
    } else
    if( strcmp( url, SETTINGS_ID ) == 0 )
    {
        const char* context = MHD_lookup_connection_value( connection
                                                         , MHD_GET_ARGUMENT_KIND
                                                         , "context"
                                                         );
        const char* username = MHD_lookup_connection_value( connection
                                                          , MHD_GET_ARGUMENT_KIND
                                                          , "username"
                                                          );
        const char* settings = MHD_lookup_connection_value( connection
                                                          , MHD_GET_ARGUMENT_KIND
                                                          , "settings"
                                                          );
        const char* create = MHD_lookup_connection_value( connection
                                                        , MHD_GET_ARGUMENT_KIND
                                                        , "create"
                                                        );
#ifdef WWW_DEBUG
    printf( "name: %s\n", username );
    printf( "settings: %s\n", settings );
    printf( "create: %s\n", create );
#endif
        data = update_settings( username, context, settings, create != NULL );
    } else
    if( strcmp( url, SUITES_ID ) == 0 )
    {
        data = get_dir_list( get_full_path(SUITES_ID), SUITES_ID );
    } else
    if( strcmp( url, TESTS_ID ) == 0 )
    {
        data = get_dir_list( get_full_path(TESTS_ID), TESTS_ID );
    } else
    if( strcmp( url, LINKS_ID ) == 0 )
    {
        data = get_dir_list( get_full_path(LINKS_ID), LINKS_ID );
    } else
    if( strcmp( url, TESTRUNS_ID ) == 0 )
    {
        data = get_dir_list( create_String( coord_getTestRunsDir() ), "prevTestRuns" );
    } else
    if( strcmp( url, START_ID ) == 0 )
    {
        const char* id = MHD_lookup_connection_value( connection
                                                    , MHD_GET_ARGUMENT_KIND
                                                    , "id"
                                                    );
        const char* name = MHD_lookup_connection_value( connection
                                                      , MHD_GET_ARGUMENT_KIND
                                                      , "name"
                                                      );
        const char* params = MHD_lookup_connection_value( connection
                                                        , MHD_GET_ARGUMENT_KIND
                                                        , "params"
                                                        );
        const char* computers = MHD_lookup_connection_value( connection
                                                           , MHD_GET_ARGUMENT_KIND
                                                           , "computers"
                                                           );
        const char* links = MHD_lookup_connection_value( connection
                                                       , MHD_GET_ARGUMENT_KIND
                                                       , "links"
                                                       );
        const char* prevTestRunId = MHD_lookup_connection_value( connection
                                                               , MHD_GET_ARGUMENT_KIND
                                                               , "prevTestRun"
                                                               );
#ifdef WWW_DEBUG
    printf( "id: %s, name: %s, params: %s, computers: %s, links: %s, prevTestRunId: %s\n"
          ,  id,     name,     params,     computers,     links,     prevTestRunId
          );
#endif

        String* testRunId = NULL;
        String* message = coord_StartTest( create_String( name )
                                         , create_String( params )
                                         , create_String( computers )
                                         , create_String( links )
                                         , prevTestRunId != NULL ? create_String( prevTestRunId ) : NULL
                                         , true // save log
                                         , &testRunId
                                         );
        StringBuffer* buf = create_StringBuffer();

        append_StringBuffer( r(buf), "{\"id\":\"" );
        append_StringBuffer( r(buf), id );
        append_StringBuffer( r(buf), "\"" );
        if( message == NULL )
        {
            append_StringBuffer( r(buf), ",\"testRunId\":\"" );
            appendString_StringBuffer( r(buf), testRunId );
            append_StringBuffer( r(buf), "\"" );
        }
        add_status( r(buf), create_String( id ), message, STATUS_STARTING );
        data = toString_Object(buf);
    } else
    if( strcmp( url, STOP_ID ) == 0 )
    {
        const char* id = MHD_lookup_connection_value( connection
                                                    , MHD_GET_ARGUMENT_KIND
                                                    , "id"
                                                    );
        const char* testRunId = MHD_lookup_connection_value( connection
                                                           , MHD_GET_ARGUMENT_KIND
                                                           , "testRunId"
                                                           );
#ifdef WWW_DEBUG
    printf( "id: %s, testRunId: %s\n", id, testRunId );
#endif
        String* message = coord_StopTest( create_String( testRunId ) );
        StringBuffer* buf = create_StringBuffer();

        append_StringBuffer( r(buf), "{\"id\":\"" );
        append_StringBuffer( r(buf), id );
        append_StringBuffer( r(buf), "\"" );
        add_status( r(buf), create_String( id ), message, STATUS_STOPPING );
        data = toString_Object(buf);
    } else
    if( strcmp( url, KILL_ID ) == 0 )
    {
        const char* id = MHD_lookup_connection_value( connection
                                                    , MHD_GET_ARGUMENT_KIND
                                                    , "id"
                                                    );
        const char* testRunId = MHD_lookup_connection_value( connection
                                                           , MHD_GET_ARGUMENT_KIND
                                                           , "testRunId"
                                                           );
#ifdef WWW_DEBUG
    printf( "id: %s, testRunId: %s\n", id, testRunId );
#endif
        String* message = KillTest_MCC( create_String( testRunId ) );
        StringBuffer* buf = create_StringBuffer();

        append_StringBuffer( r(buf), "{\"id\":\"" );
        append_StringBuffer( r(buf), id );
        append_StringBuffer( r(buf), "\"" );
        add_status( r(buf), create_String( id ), message, STATUS_KILLING );
        data = toString_Object(buf);
    } else
    if( strcmp( url, INFO_ID ) == 0 )
    {
        data = coord_GetInfoAll();
    } else
    if( strcmp( url, TEST_ID ) == 0 )
    {
        const char* testRunId = MHD_lookup_connection_value( connection
                                                           , MHD_GET_ARGUMENT_KIND
                                                           , "testRunId"
                                                           );
        const char* index = MHD_lookup_connection_value( connection
                                                       , MHD_GET_ARGUMENT_KIND
                                                       , "index"
                                                       );
        String* message = coord_GetNodeInfo( create_String(testRunId), atoi(index) );
        StringBuffer* buf = create_StringBuffer();

        append_StringBuffer( r(buf), "{\"id\":\"" );
        append_StringBuffer( r(buf), testRunId );
        append_StringBuffer( r(buf), "\",\"index\":" );
        append_StringBuffer( r(buf), index );
        if( message != NULL )
        {
            append_StringBuffer( r(buf), ",\"info\":" );
            appendString_StringBuffer( r(buf), message );
        }
        add_status( r(buf), create_String( testRunId ), NULL, STATUS_OK );
        data = toString_Object(buf);
    } else
    if( strcmp( url, SAVE_ID ) == 0 )
    {
        const char* context = MHD_lookup_connection_value( connection
                                                         , MHD_GET_ARGUMENT_KIND
                                                         , "context"
                                                         );
        const char* name = MHD_lookup_connection_value( connection
                                                      , MHD_GET_ARGUMENT_KIND
                                                      , "name"
                                                      );
        const char* tests = MHD_lookup_connection_value( connection
                                                       , MHD_GET_ARGUMENT_KIND
                                                       , "tests"
                                                       );
#ifdef WWW_DEBUG
    printf( "name: %s, tests: %s\n", name, tests );
#endif
        data = save_tests( name, tests, context );
    } else
    if( strcmp( url, LOAD_ID ) == 0 )
    {
        const char* context = MHD_lookup_connection_value( connection
                                                         , MHD_GET_ARGUMENT_KIND
                                                         , "context"
                                                         );
        const char* name = MHD_lookup_connection_value( connection
                                                      , MHD_GET_ARGUMENT_KIND
                                                      , "name"
                                                      );
#ifdef WWW_DEBUG
    printf( "name: %s\n", name );
#endif
        data = load_tests( name, context );
    } else {
        bindata = get_file( url, &size );
        content_type = 0;
#ifdef WWW_DEBUG
        if( bindata == NULL ) printf( "not found url=%s\n", url );
#endif
    }

    if( data )
    {
        bindata = toCharArray_String( r(data) );
        size = length_String( r(data) );
#ifdef WWW_DEBUG
        printf( "answer: %s\n", toCharArray_String( r(data) ) );
#endif
    }

    return respond( connection, (char*)bindata, size, content_type );
}


static int answer_to_connection( void *cls
                               , struct MHD_Connection *connection
                               , const char *url
                               , const char *method
                               , const char *version
                               , const char *upload_data
                               , size_t *upload_data_size
                               , void **ptr
                               )
{
#ifdef WWW_DEBUG
    printf( "url=%s\n", url );
#endif

    if( 0 == strcmp( method, "GET" ) )
    {
        return process_get_request( connection, url );
    } else {
        /* unexpected method */
        return MHD_NO;
    }
}

int main( int argc, char** argv )
{
    int port = COORD_PORT;
    const char* cluster;
    const char* config_dir;
    const char* testruns_dir;
    String* res;

    printf( "httpmcc port conf_dir cluster testrun_dir\n" );

    if( argc > 1 )
    {
        port = atoi( argv[1] );
    }

    res = coord_Init( argc > 2 ? argv[2] : NULL
                    , argc > 3 ? argv[3] : NULL
                    , argc > 4 ? argv[4] : NULL
                    );
    if( res != NULL )
    {
        puts( toCharArray_String( r(res) ) );
        destroy( res );
        return -1;
    }

    cluster = coord_getCluster();
    config_dir = coord_getConfigDir();
    testruns_dir = coord_getTestRunsDir();

    printf( "httpmcc port='%d' cluster='%s' conf_dir='%s' testruns_dir='%s'\n"
          , port
          , cluster
          , config_dir
          , testruns_dir
          );

    struct MHD_Daemon *daemon = MHD_start_daemon( MHD_USE_SELECT_INTERNALLY
                                                , port
                                                , NULL
                                                , NULL
                                                , &answer_to_connection
                                                , NULL
                                                , MHD_OPTION_END
                                                );
    if( NULL == daemon )
    {
        perror( "can't create web server" );
        return 1;
    }

    getchar ();

    MHD_stop_daemon( daemon );

    return 0;
}
